#include <stdio.h>
#include <tcl.h>

#include "includes.h"

#define CTABLE_SIZE 100

static int SmbCloseProc(ClientData instanceData,Tcl_Interp *interp);
static int SmbGetHandleProc(ClientData instanceData,int direction,ClientData *handlePtr);
static int SmbInputProc(ClientData instanceData,char *buf,int bufSize,int *errorCodePtr);
static int SmbOutputProc(ClientData instanceData,char *buf,int toWrite,int *errorCodePtr);
static void SmbWatchProc(ClientData instanceData,int mask);
static int SmbSeekProc(ClientData instanceData,long offset,int mode,int *errorCodePtr);

static Tcl_ChannelType SmbChannelType = {
  "samba",			/* Type name */
  NULL,				/* Set blocking/nonblocking mode */
  SmbCloseProc,			/* Close proc */
  SmbInputProc,			/* Input proc */
  SmbOutputProc,		/* Output proc */
  SmbSeekProc,			/* Seek proc */
  NULL,				/* Set option proc */
  NULL,				/* Get option proc */
  SmbWatchProc,			/* Initialize notifier */
  SmbGetHandleProc,		/* Get handles out of channel */
};
  
struct credentials {
  char hostname[14],hostip[14],workgroup[14],share[255],password[255],username[14],myname[14],dev[5];
  int max_protocol;
};

typedef struct SmbChannel {
  Tcl_Channel channel;
  int flags;
  int handle;
  int fd;
  int offset;
} SmbChannel;

struct Smb_Connection {
  struct cli_state c;
  struct credentials creds;
  int refcount;
  int used;
};

struct Smb_Connection ctable[CTABLE_SIZE];

int ServerFD=-1;
extern struct in_addr ipzero;
extern pstring scope;
extern int DEBUGLEVEL;

static BOOL open_sockets();
int nmbcmd(ClientData,Tcl_Interp *,int argc,char **argv);
int smbcmd(ClientData,Tcl_Interp *,int argc,char **argv);

int Samba_Init(Tcl_Interp *interp) {
  static pstring servicesf = CONFIGFILE;  
  int i;

  Tcl_CreateCommand(interp,"nmb",nmbcmd,(ClientData)0,(Tcl_CmdDeleteProc *)0);
  Tcl_CreateCommand(interp,"smb",smbcmd,(ClientData)0,(Tcl_CmdDeleteProc *)0);

  Tcl_PkgProvide(interp,"samba","2.0");

  for (i=0;i<CTABLE_SIZE;i++) {
    ctable[i].used=0;
  }

  DEBUGLEVEL=0;

  TimeInit();
  charset_initialise();
  setup_logging("samba-tcl",True);
  
  if (!lp_load(servicesf,True)) {
    Tcl_AppendResult(interp,"Can't load...",(char *)NULL);
  }

  load_interfaces();

  if (open_sockets()==TCL_ERROR) {
    Tcl_AppendResult(interp,"Could not open sockets",(char *)NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

static void _interpret_node_status(Tcl_Interp *interp,char *p, char *master,char *rname) {
  char buf[100];
  int numnames = CVAL(p,0);

  if (rname) *rname = 0;
  if (master) *master = 0;

  p += 1;
  while (numnames--) {
    char qname[17];
    int type;
    fstring flags;
    int i;
    *flags = 0;
    StrnCpy(qname,p,15);
    type = CVAL(p,15);
    p += 16;

    if (master && !*master && type == 0x1d) {
      StrnCpy(master,qname,15);
      trim_string(master,NULL," ");
    }

    if (rname && !*rname && type == 0x20 && !(p[0]&0x80)) {
      StrnCpy(rname,qname,15);
      trim_string(rname,NULL," ");
    }
      

    snprintf(buf,100,"{ {%s} {%u} {%u} } ",qname,type,p[0]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    p+=2;
  }
/*
  printf("%d %d\n",IVAL(p,20),IVAL(p,24));
*/
}

int Nmb_Status(Tcl_Interp *interp,int argc,char **argv) {
  struct in_addr to_ip;
  char *name;
  int name_type;
  BOOL recurse=True;
  char *master=NULL;
  char *rname=NULL;
  void (*fn)(struct packet_struct *)=NULL;
  BOOL found=False;
  int retries = 2;
  int retry_time = 5000;
  struct timeval tval;
  struct packet_struct p;
  struct packet_struct *p2;
  struct nmb_packet *nmb = &p.packet.nmb;
  static int name_trn_id = 0;

  if (argc<4) {
    Tcl_AppendResult(interp,"Usage: nmb status name type ip",(char *)NULL);
    return TCL_ERROR;
  }
  name=strdup(argv[1]);
  name_type=atoi(argv[2]);
  to_ip=*interpret_addr2(argv[3]);

  bzero((char *)&p,sizeof(p));

  if (!name_trn_id) name_trn_id = (time(NULL)%(unsigned)0x7FFF) + 
    (getpid()%(unsigned)100);
  name_trn_id = (name_trn_id+1) % (unsigned)0x7FFF;

  nmb->header.name_trn_id = name_trn_id;
  nmb->header.opcode = 0;
  nmb->header.response = False;
  nmb->header.nm_flags.bcast = False;
  nmb->header.nm_flags.recursion_available = False;
  nmb->header.nm_flags.recursion_desired = False;
  nmb->header.nm_flags.trunc = False;
  nmb->header.nm_flags.authoritative = False;
  nmb->header.rcode = 0;
  nmb->header.qdcount = 1;
  nmb->header.ancount = 0;
  nmb->header.nscount = 0;
  nmb->header.arcount = 0;

  make_nmb_name(&nmb->question.question_name,name,name_type,scope);

  nmb->question.question_type = 0x21;
  nmb->question.question_class = 0x1;

  p.ip = to_ip;
  p.port = NMB_PORT;
  p.fd = ServerFD;
  p.timestamp = time(NULL);
  p.packet_type = NMB_PACKET;

  GetTimeOfDay(&tval);

  if (!send_packet(&p)) {
    Tcl_AppendResult(interp,"Could not send packet",(char *)NULL);
    return TCL_ERROR;
  }

  retries--;

  while (1)
    {
      struct timeval tval2;
      GetTimeOfDay(&tval2);
      if (TvalDiff(&tval,&tval2) > retry_time) {
	if (!retries) break;
	if (!found && !send_packet(&p)) {
          Tcl_AppendResult(interp,"Could not send packet",(char *)NULL);
	  return TCL_ERROR;
	}
	GetTimeOfDay(&tval);
	retries--;
      }

      if ((p2=receive_packet(ServerFD,NMB_PACKET,90))) {
	  struct nmb_packet *nmb2 = &p2->packet.nmb;
          debug_nmb_packet(p2);

	  if (nmb->header.name_trn_id != nmb2->header.name_trn_id ||
	      !nmb2->header.response) {
	    /* its not for us - maybe deal with it later */
	    if (fn) 
	      fn(p2);
	    else
	      free_packet(p2);
	    continue;
	  }
	  
	  if (nmb2->header.opcode != 0 ||
	      nmb2->header.nm_flags.bcast ||
	      nmb2->header.rcode ||
	      !nmb2->header.ancount ||
	      nmb2->answers->rr_type != 0x21) {
	    /* XXXX what do we do with this? could be a redirect, but
	       we'll discard it for the moment */
	    free_packet(p2);
	    continue;
	  }

	  _interpret_node_status(interp,&nmb2->answers->rdata[0], master,rname);
	  free_packet(p2);
	  return TCL_OK;
	}
    }
  
  Tcl_AppendResult(interp,"{ {NORESPONSE} {0} {0} }",(char *)NULL);
  return TCL_OK;
}

int Nmb_Lookup(Tcl_Interp *interp,int argc,char **argv) {
  char *lookup;
  int lookup_type;
  BOOL use_bcast;
  BOOL recursion_desired;
  struct in_addr bcast_addr;
  struct in_addr *ip_list;
  int count,j;

  if (argc<5) {
    Tcl_AppendResult(interp,"Usage: nmb lookup name type bcast recursion ?addr?",(char *)NULL);
    return TCL_ERROR;
  }

  lookup=strdup(argv[1]);
  lookup_type=atoi(argv[2]);
  use_bcast=atoi(argv[3]);
  recursion_desired=atoi(argv[4]);
  
  if (argc>5) {
    bcast_addr=*interpret_addr2(argv[5]);
  } else {
    bcast_addr=*iface_bcast(ipzero);
  }

  if ((ip_list=name_query(ServerFD,lookup,lookup_type,use_bcast,
                          recursion_desired,bcast_addr,&count,NULL))) {
    Tcl_AppendResult(interp,inet_ntoa(ip_list[0]),(char *)NULL);
    for (j=1;j<count;j++) {
      Tcl_AppendResult(interp," ",inet_ntoa(ip_list[j]),(char *)NULL);
    }
    free(lookup);
    return TCL_OK;
  } else {
    Tcl_AppendResult(interp,"Could not find name",(char *)NULL);
    return TCL_ERROR;
  }
}

static int open_connection(Tcl_Interp *interp,struct cli_state *c,struct credentials *creds) {
  if (!cli_initialise(c) || !cli_connect(c, creds->hostip, NULL)) {
    Tcl_AppendResult(interp,"Failed to connect with ", creds->hostip,(char *)NULL);
    return TCL_ERROR;
  }

  if (!cli_session_request(c, creds->hostname, 0x20, creds->myname)) {
    Tcl_AppendResult(interp, creds->hostname," rejected the session",(char *)NULL);
    cli_shutdown(c);
    return TCL_ERROR;
  }

  c->protocol = creds->max_protocol;

  if (!cli_negprot(c)) {
    Tcl_AppendResult(interp,creds->hostname," rejected the negprot (",cli_errstr(c),")",(char *)NULL);
    cli_shutdown(c);
    return TCL_ERROR;
  }

  if (!cli_session_setup(c, creds->username, creds->password, strlen(creds->password),
                         "", 0, creds->workgroup)) {
    Tcl_AppendResult(interp, creds->hostname," rejected the sessionsetup (",cli_errstr(c),")",(char *)NULL);
    cli_shutdown(c);
    return TCL_ERROR;
  }

  strncpy(creds->dev,"A:",5);
  if (strstr(creds->share,"IPC$")) strncpy(creds->dev,"IPC",5);

  if (!cli_send_tconX(c, creds->share, creds->dev, creds->password, strlen(creds->password)+1)) {
    Tcl_AppendResult(interp, creds->hostname," refused tree connect (",cli_errstr(c),")",(char *)NULL);
    cli_shutdown(c);
    return TCL_ERROR;
  }

  return TCL_OK;
}

static int close_connection(Tcl_Interp *interp,struct cli_state *c) {
  if (!cli_tdis(c)) {
    Tcl_AppendResult(interp,"tdis failed (",cli_errstr(c),")\0");
    return TCL_ERROR;
  }

  cli_shutdown(c);
  return TCL_OK;
}

static int check_handle(Tcl_Interp *interp,int i) {
  if (i<0||i>=CTABLE_SIZE) {
    Tcl_AppendResult(interp,"Handle out of range.",(char *)NULL);
    return TCL_ERROR;
  }
  if (ctable[i].used==0) {
    Tcl_AppendResult(interp,"Handle not in use.",(char *)NULL);
    return TCL_ERROR;
  }
}

int Smb_Connect(Tcl_Interp *interp,int argc,char **argv) {
  struct credentials creds;
  struct cli_state c;
  int i,ret;
  char buf[100];

  if (argc!=9) {
    Tcl_AppendResult(interp,"Usage: smb connect hostname ip workgroup share username password localhostname max_protocol",(char *)NULL);
    return TCL_ERROR;
  }
  
  for (i=0;i<CTABLE_SIZE;i++) {
    if (ctable[i].used==0) break;
  }
  
  if (i==CTABLE_SIZE) {
    Tcl_AppendResult(interp,"No free entries in the connection table",(char *)NULL);
    return TCL_ERROR;
  }
  
  strncpy(ctable[i].creds.hostname,argv[1],14);
  strncpy(ctable[i].creds.hostip,argv[2],14);
  strncpy(ctable[i].creds.workgroup,argv[3],14);
  strncpy(ctable[i].creds.share,argv[4],255);
  strncpy(ctable[i].creds.password,argv[6],255);
  strncpy(ctable[i].creds.username,argv[5],14);
  strncpy(ctable[i].creds.myname,argv[7],14);
  if (strcmp(argv[8],"CORE")==0) {
    ctable[i].creds.max_protocol=PROTOCOL_CORE;
  } else if (strcmp(argv[8],"COREPLUS")==0) {
    ctable[i].creds.max_protocol=PROTOCOL_COREPLUS;
  } else if (strcmp(argv[8],"LANMAN1")==0) {
    ctable[i].creds.max_protocol=PROTOCOL_LANMAN1;
  } else if (strcmp(argv[8],"LANMAN2")==0) {
    ctable[i].creds.max_protocol=PROTOCOL_LANMAN2;
  } else if (strcmp(argv[8],"NT1")==0) {
    ctable[i].creds.max_protocol=PROTOCOL_NT1;
  } else {
    Tcl_AppendResult(interp,"Unknown max_protocol: ",argv[8],". Must be one of CORE,COREPLUS,LANMAN1,LANMAN2,NT1",(char *)NULL);
    return TCL_ERROR;
  }

  ret=open_connection(interp,&ctable[i].c,&ctable[i].creds);
  if (ret==TCL_ERROR) return TCL_ERROR;
  ctable[i].refcount=0;
  ctable[i].used=1;
  snprintf(buf,100,"%d",i);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int Smb_Disconnect(Tcl_Interp *interp,int argc,char **argv) {
  int i;

  if (argc!=2) {
    Tcl_AppendResult(interp,"Usage: smb close handle",(char *)NULL);
    return TCL_ERROR;
  }
  i=atoi(argv[1]);
  if (check_handle(interp,i)==TCL_ERROR) return TCL_ERROR;
  if (ctable[i].refcount>0) {
    Tcl_AppendResult(interp,"The handle you are trying to close is in use.",(char *)NULL);
    return TCL_ERROR;
  }
  close_connection(interp,&ctable[i].c);
  ctable[i].used=0;
  return TCL_OK;
}

int Smb_Browse(Tcl_Interp *interp,int argc,char **argv) {
  int i;

  void browse_callback(char *sname,uint32 stype,char *comment) {
    char buf[1000];
    
    snprintf(buf,1000," { {%s} {%08x} {%s} }",sname,stype,comment);
    Tcl_AppendResult(interp,buf,(char *)NULL);
  } 

  if (argc!=3) {
    Tcl_AppendResult(interp,"Usage: smb browse type handle",(char *)NULL);
    return TCL_ERROR;
  }
  i=atoi(argv[2]);
  if (check_handle(interp,i)==TCL_ERROR) return TCL_ERROR;
  if (strcmp(argv[1],"net")==0) {
    cli_NetServerEnum(&ctable[i].c,ctable[i].creds.workgroup,SV_TYPE_ALL,browse_callback);
  } else if (strcmp(argv[1],"shares")==0) {
    cli_RNetShareEnum(&ctable[i].c,browse_callback);
  }
    
  return TCL_OK;
}

int Smb_Open(Tcl_Interp *interp,int argc,char **argv) {
  int handle,fd,i,flags;
  char channelName[20];
  SmbChannel *smbchan;

  if (argc<3||argc>4) {
    Tcl_AppendResult(interp,"Usage: smb openchannel handle filename ?mode?",(char *)NULL);
    return TCL_ERROR;
  }
  
  handle=atoi(argv[1]);
  if (check_handle(interp,handle)==TCL_ERROR) return TCL_ERROR;

  if (argc==4) {
    if (strcmp(argv[3],"r")==0) {
      flags=O_RDONLY;
    } else if (strcmp(argv[3],"r+")==0) {
      flags=O_RDWR;
    } else if (strcmp(argv[3],"w")==0) {
      Tcl_AppendResult(interp,"Write only not supported",(char *)NULL);
      return TCL_ERROR;
    } else if (strcmp(argv[3],"w+")==0) {
      flags=O_RDWR|O_CREAT|O_TRUNC;
    } else if (strcmp(argv[3],"a")==0) {
      Tcl_AppendResult(interp,"Append not supported",(char *)NULL);
      return TCL_ERROR;
    } else if (strcmp(argv[3],"a+")==0) {
      Tcl_AppendResult(interp,"Append+ not supported",(char *)NULL);
      return TCL_ERROR;
    }
  } else {
    flags=O_RDONLY;
  }
    

  fd=cli_open(&ctable[handle].c,argv[2],flags,DENY_NONE);
  if (fd==-1) {
    Tcl_AppendResult(interp,"Could not open ",argv[2]," (",cli_errstr(&ctable[handle].c),").",(char *)NULL);
    return TCL_ERROR;
  }

  smbchan=(SmbChannel *)ckalloc((unsigned)sizeof(SmbChannel));

  smbchan->handle=handle;
  smbchan->fd=fd;
  smbchan->offset=0;

  snprintf(channelName,20,"smb%d_%d",smbchan->handle,smbchan->fd);
  
  smbchan->channel=Tcl_CreateChannel(&SmbChannelType,channelName,
    (ClientData)smbchan,TCL_READABLE|TCL_WRITABLE);
    
  ctable[handle].refcount++;

  Tcl_RegisterChannel(interp,smbchan->channel);
  Tcl_AppendResult(interp,channelName,(char *)NULL); 
  return TCL_OK;
}

static int SmbCloseProc(ClientData instanceData,Tcl_Interp *interp) {
  SmbChannel *smbchan=(SmbChannel *)instanceData;
  int handle,fd;
  
  handle=smbchan->handle;
  fd=smbchan->fd;

  if (check_handle(interp,handle)==TCL_ERROR) return TCL_ERROR;
  if (!cli_close(&ctable[handle].c,fd)) {
    return TCL_ERROR; 
  }
  ckfree((char *)smbchan);
  ctable[handle].refcount--;
  return TCL_OK;
}

static int SmbGetHandleProc(ClientData instanceData,int direction,ClientData *handlePtr) {
  SmbChannel *smbchan=(SmbChannel *)instanceData;

  *handlePtr=(ClientData)smbchan->fd;
  return TCL_OK;  
}

static int SmbOutputProc(ClientData instanceData,char *buf,int toWrite,int *errorCodePtr) {
  SmbChannel *smbchan=(SmbChannel *)instanceData;
  int handle,fd,size;

  handle=smbchan->handle;
  fd=smbchan->fd;

  size=cli_write(&ctable[handle].c,fd,buf,smbchan->offset,toWrite);
  if (size<0) return -1;
  smbchan->offset+=size;
  return size;
}

static int SmbInputProc(ClientData instanceData,char *buf,int bufSize,int *errorCodePtr) {
  SmbChannel *smbchan=(SmbChannel *)instanceData;
  int handle,fd,size;

  handle=smbchan->handle;
  fd=smbchan->fd;

  size=cli_read(&ctable[handle].c,fd,buf,smbchan->offset,bufSize);
  if (size<0) return -1;
  smbchan->offset+=size;
  return size;
}

static void SmbWatchProc(ClientData instanceData,int mask) {
  SmbChannel *smbchan=(SmbChannel *)instanceData;
  
  if (mask) {
    Tcl_CreateFileHandler(smbchan->fd,mask,(Tcl_FileProc *)Tcl_NotifyChannel,
      (ClientData)smbchan->channel);
  } else {
    Tcl_DeleteFileHandler(smbchan->fd);
  }
}

static int SmbSeekProc(ClientData instanceData,long offset,int mode,int *errorCodePtr) {
  SmbChannel *smbchan=(SmbChannel *)instanceData;
  
  if (mode==SEEK_SET) smbchan->offset=offset;
  else if (mode==SEEK_CUR) smbchan->offset+=offset;
  else if (mode==SEEK_END) return TCL_ERROR;
  else return TCL_ERROR;

  if (smbchan->offset<0) smbchan->offset=0;
  return TCL_OK;
}  

int Smb_Mv(Tcl_Interp *interp,int argc,char **argv) {
  int i;

  if (argc!=4) {
    Tcl_AppendResult(interp,"Usage: smb mv handle source dest",(char *)NULL);
    return TCL_ERROR;
  }
  i=atoi(argv[1]);
  if (check_handle(interp,i)==TCL_ERROR) return TCL_ERROR;
  if (!cli_mv(&ctable[i].c,argv[2],argv[3])) {
    Tcl_AppendResult(interp,"Could not move ",argv[2]," to ",argv[3]," (",cli_errstr(&ctable[i].c),").",(char *)NULL);
    return TCL_ERROR;
  }
  ctable[i].refcount--;
  return TCL_OK;
}

int Smb_Mkdir(Tcl_Interp *interp,int argc,char **argv) {
  int i;

  if (argc!=3) {
    Tcl_AppendResult(interp,"Usage: smb mkdir handle dirname",(char *)NULL);
    return TCL_ERROR;
  }
  i=atoi(argv[1]);
  if (check_handle(interp,i)==TCL_ERROR) return TCL_ERROR;
  if (!cli_mkdir(&ctable[i].c,argv[2])) {
    Tcl_AppendResult(interp,"Could not make directory ",argv[2]," (",cli_errstr(&ctable[i].c),").",(char *)NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

int Smb_Rmdir(Tcl_Interp *interp,int argc,char **argv) {
  int i;

  if (argc!=3) {
    Tcl_AppendResult(interp,"Usage: smb rmdir handle dirname",(char *)NULL);
    return TCL_ERROR;
  }
  i=atoi(argv[1]);
  if (check_handle(interp,i)==TCL_ERROR) return TCL_ERROR;
  if (!cli_rmdir(&ctable[i].c,argv[2])) {
    Tcl_AppendResult(interp,"Could not remove directory ",argv[2]," (",cli_errstr(&ctable[i].c),").",(char *)NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

int Smb_Unlink(Tcl_Interp *interp,int argc,char **argv) {
  int i,fd;

  if (argc!=3) {
    Tcl_AppendResult(interp,"Usage: smb unlink handle filename",(char *)NULL);
    return TCL_ERROR;
  }
  i=atoi(argv[1]);
  if (check_handle(interp,i)==TCL_ERROR) return TCL_ERROR;
  if (!cli_unlink(&ctable[i].c,argv[2])) {
    Tcl_AppendResult(interp,"Could not unlink ",argv[2]," (",cli_errstr(&ctable[i].c),").",(char *)NULL); 
   return TCL_ERROR;
  }
  return TCL_OK;
}

int Smb_Ls(Tcl_Interp *interp,int argc,char **argv) {
  int i;

  int dir_callback(file_info *finfo) {
    char buf[1000];
    
    snprintf(buf,1000," { {%s} {%lu} {%08x} {%lu} {%lu} {%lu} {%lu} {%lu} }",
             finfo->name,finfo->size,finfo->mode,finfo->uid,finfo->gid,
             finfo->mtime,finfo->atime,finfo->ctime);
    Tcl_AppendResult(interp,buf,(char *)NULL);
  }

  if (argc!=3) {
    Tcl_AppendResult(interp,"Usage: smb dir handle mask",(char *)NULL);
    return TCL_ERROR;
  }
  i=atoi(argv[1]);
  if (check_handle(interp,i)==TCL_ERROR) return TCL_ERROR;
  
  if (!cli_find(&ctable[i].c,argv[2],dir_callback)) {
    Tcl_AppendResult(interp,"Could not list directory (",cli_errstr(&ctable[i].c),").",(char *)NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

int Smb_Dump(Tcl_Interp *interp,int argc,char **argv) { 
  int i;
  char buf[100];
  
  for (i=0;i<CTABLE_SIZE;i++) {
    if (ctable[i].used==1) {
      snprintf(buf,80,"%d:\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%d\n",
               i,ctable[i].creds.hostname,
               ctable[i].creds.hostip,
               ctable[i].creds.workgroup,
               ctable[i].creds.share,
               ctable[i].creds.password,
               ctable[i].creds.username,
               ctable[i].creds.myname,
               ctable[i].creds.max_protocol);
      Tcl_AppendResult(interp,buf,(char *)NULL);
    }
  }
  return TCL_OK;
}

int open_sockets() {
  ServerFD=open_socket_in(SOCK_DGRAM,137,3,
                          interpret_addr(lp_socket_address()));

  if (ServerFD==-1) return TCL_ERROR;
  
  set_socket_options(ServerFD,"SO_BROADCAST");

  return TCL_OK;
}

int nmbcmd(ClientData unused,Tcl_Interp *interp,int argc,char **argv) {
  if (argc<2) {
    Tcl_AppendResult(interp,"Usage: nmb option ?args?",(char *)NULL);
    return TCL_ERROR;
  }
  if (strcmp(argv[1],"lookup")==0) {
    return Nmb_Lookup(interp,argc-1,argv+1);
  } else if (strcmp(argv[1],"status")==0) {
    return Nmb_Status(interp,argc-1,argv+1);
  } else {
    Tcl_AppendResult(interp,"Usage: nmb [lookup|status] ?args?",(char *)NULL);
    return TCL_ERROR;
  }  
}

int smbcmd(ClientData unused,Tcl_Interp *interp,int argc,char **argv) {
  if (argc<2) {
    Tcl_AppendResult(interp,"Usage: smb option ?args?",(char *)NULL);
    return TCL_ERROR;
  }
  if (strcmp(argv[1],"connect")==0) {
    return Smb_Connect(interp,argc-1,argv+1);
  } else if (strcmp(argv[1],"disconnect")==0) {
    return Smb_Disconnect(interp,argc-1,argv+1);
  } else if (strcmp(argv[1],"browse")==0) {
    return Smb_Browse(interp,argc-1,argv+1);
  } else if (strcmp(argv[1],"ls")==0) {
    return Smb_Ls(interp,argc-1,argv+1);
  } else if (strcmp(argv[1],"mv")==0) {
    return Smb_Mv(interp,argc-1,argv+1);
  } else if (strcmp(argv[1],"mkdir")==0) {
    return Smb_Mkdir(interp,argc-1,argv+1);
  } else if (strcmp(argv[1],"unlink")==0) {
    return Smb_Unlink(interp,argc-1,argv+1);
  } else if (strcmp(argv[1],"rmdir")==0) {
    return Smb_Rmdir(interp,argc-1,argv+1);
  } else if (strcmp(argv[1],"open")==0) {
    return Smb_Open(interp,argc-1,argv+1);
  } else if (strcmp(argv[1],"dump")==0) {
    return Smb_Dump(interp,argc-1,argv+1);
  } else {
    Tcl_AppendResult(interp,"Usage: smb [dump|connect|disconnect|ls|mv|mkdir|unlink|rmdir|browse] ?args?",(char *)NULL);
    return TCL_ERROR;
  }  
}
